
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#ifdef	NEED_NETINET_IN_H
#include <netinet/in.h>		/* ntohl() */
#else	/* NEED_NETINET_IN_H */
#include <sys/param.h>		/* ntohl() */
#endif	/* NEED_NETINET_IN_H */

#include <erl_driver.h>
#include <erl_driver_tk.h>

/*
** An 'ErlDrvBinary' is slightly smaller than a 'Binary', which is
** what the Erlang VM is expecting to use.  However, the VM won't
** be touching anything we make, so we don't have to worry about
** the additional hidden fields in a 'Binary' structure.
*/

static ErlIOVec _ev;
static char *_bufp = NULL;
static int _bufsiz = 32 * 1024;
static struct iovec _iov[2];
static ErlDrvBinary *_binv[2];

ErlIOVec *
read_packet_make_iovec(int fd)
{
    char	lenbuf[4];
    int		wanted, numread, got;

    if (_bufp == NULL) {
	if ((_bufp = edtk_driver_alloc_wrapper(_bufsiz)) == NULL) {
	    edtk_debug("%s: malloc failed", __FUNCTION__);
	    return NULL;
	}
    }
    wanted = 4;
    got = 0;
    while (got < wanted) {
	if ((numread = read(fd, lenbuf + got, wanted - got)) < 0) {
	    edtk_debug("%s: read 1 failed, errno = %d", __FUNCTION__, errno);
	    return NULL;
	} else if (numread == 0) {
	    edtk_debug("%s: 1 EOF", __FUNCTION__);
	    return NULL;
	}
	got += numread;
    }
    wanted = ntohl(*(unsigned long *) lenbuf);
    edtk_debug("%s: wanted = %lu", __FUNCTION__, wanted);
    if (wanted > _bufsiz) {
	edtk_driver_free_wrapper(_bufp);
	_bufp = NULL;
	_bufsiz = wanted + _bufsiz;
	return read_packet_make_iovec(fd);
    }
    got = 0;
    while (got < wanted) {
	edtk_debug("%s: top got < wanted loop", __FUNCTION__);
	if ((numread = read(fd, _bufp + got, wanted - got)) < 0) {
	    edtk_debug("%s: read 2 failed, errno = %d", __FUNCTION__, errno);
	    return NULL;
	} else if (numread == 0) {
	    edtk_debug("%s: 2 EOF", __FUNCTION__);
	    return NULL;
	}
	got += numread;
	edtk_debug("%s: got = %lu", __FUNCTION__, got);
    }

    _iov[0].iov_base = NULL;
    _iov[0].iov_len = 0;
    _iov[1].iov_base = _bufp;
    _iov[1].iov_len = wanted;
    _binv[0] = NULL;
    _binv[1] = edtk_alloced_ptr2ErlDrvBinary(_bufp);
    _binv[1]->orig_size = _bufsiz;
    _binv[1]->refc = 1;
    /* Now we've got everything ready for _ev */
    _ev.vsize = 2;
    _ev.size = wanted;
    _ev.iov = (SysIOVec *) &_iov;
    _ev.binv = _binv;
    return &_ev;
}

/*
** Provide a work-alike versions for driver_*_binary().
*/

ErlDrvBinary *
driver_alloc_binary(int size)
{
    ErlDrvBinary        *ebin;

    if ((ebin = malloc(size + sizeof(ErlDrvBinary))) != NULL) {
        ebin->refc = 1;
        ebin->orig_size = size;
    }
    return ebin;
}

ErlDrvBinary *
driver_realloc_binary(ErlDrvBinary *oldbin, int size)
{
    ErlDrvBinary        *ebin = NULL;

    if (oldbin != NULL &&
        (ebin = realloc(oldbin, size + sizeof(ErlDrvBinary))) != NULL) {
        ebin->orig_size = size;
    }
    return ebin;
}

void
driver_free_binary(ErlDrvBinary *ebin)
{
    if (ebin != NULL) {
        if (ebin->refc == 1) {
            free(ebin);
        } else {
            ebin->refc--;
        }
    }
}

/*
** {sigh} erl_driver_tk.c refers to sys_{alloc,free}().
*/

void *
sys_alloc(size_t size)
{
    return malloc(size);
}

void
sys_free(void *p)
{
    if (p != NULL) {
	free(p);
    }
}

int
driver_output_term(ErlDrvPort ix, ErlDrvTermData* data, int len)
{
    ErlDrvTermData	*ptr, *ptr_end, tag;
    int			bytes;
    char		bytesbuf[4];
    static char		*buf = NULL, *bufp;
    static int		bufsiz = 32768;
    int			i, tmpi;
    ErlDrvBinary	*ebin;

    edtk_debug("%s: stub is in-progress\r\n", __FUNCTION__);

    if (buf == NULL) {
	if ((buf = (char *) malloc(bufsiz)) == NULL) {
	    return -1;
	}
    }

    bytes = 0;
    ptr = data;
    ptr_end = ptr + len;
    while (ptr < ptr_end) {
	tag = *ptr++;
	switch (tag) {
	case ERL_DRV_NIL:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: nil\r\n", __FUNCTION__);
	    break;
	case ERL_DRV_ATOM:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: atom = %s\r\n", __FUNCTION__, (char *) *ptr);
	    bytes++;		/* name len, limited to 255! */
	    i = strlen((char *) *ptr);
	    edtk_debug("%s: bc: i = %d\r\n", __FUNCTION__, i);
	    bytes += i;
	    ptr++;
	    break;
	case ERL_DRV_PORT:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: port = 0x%lx\r\n", __FUNCTION__, *ptr);
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_INT:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: int = %ld\r\n", __FUNCTION__, *ptr);
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_TUPLE:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: tuple size = %ld\r\n", __FUNCTION__, *ptr);
	    bytes++;		/* size of tuple, limited to 255! */
	    ptr++;
	    break;
	case ERL_DRV_LIST:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: list size = %ld\r\n", __FUNCTION__, *ptr);
	    bytes += 4;
	    ptr++;
	    break;
	case ERL_DRV_BINARY:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: binary/string 0x%lx + 0x%lx size = %ld\r\n", __FUNCTION__, ptr[0], ptr[2], ptr[1]);
	    bytes += 4;		/* Length indicator */
	    bytes += ptr[1];
	    ptr += 3;
	    break;
	case ERL_DRV_STRING:
	    bytes++;		/* Tag */
	    edtk_debug("%s: bc: binary/string 0x%lx size = %ld\r\n", __FUNCTION__, ptr[0], ptr[1]);
	    bytes += 4;		/* Length indicator */
	    bytes += ptr[1];
	    ptr += 2;
	    break;
	default:
	    edtk_debug("%s: bc: UNKNOWN tag = %d\r\n", __FUNCTION__, tag);
	    break;
	}
	edtk_debug("%s: bc: bytes = %ld\r\n", __FUNCTION__, bytes);
    }
    edtk_debug("%s: bc: total = %ld\r\n", __FUNCTION__, bytes);

    if (bytes > bufsiz) {
	free(buf);
	bufsiz += bytes;
	return driver_output_term(ix, data, len); /* Lazy */
    }

    bufp = buf;
    ptr = data;
    ptr_end = ptr + len;
    while (ptr < ptr_end) {
	tag = *ptr++;
	switch (tag) {
	case ERL_DRV_NIL:
	    *bufp++ = PIPE_DRIVER_TERM_NIL;
	    break;
	case ERL_DRV_ATOM:
	    *bufp++ = PIPE_DRIVER_TERM_ATOM;
	    i = strlen((char *) *ptr);
	    *bufp++ = i;
	    memcpy(bufp, (char *) *ptr, i);
	    bufp += i;
	    ptr++;
	    break;
	case ERL_DRV_PORT:
	    *bufp++ = PIPE_DRIVER_TERM_PORT;
	    *(unsigned long *) bytesbuf = htonl(*ptr);
	    memcpy(bufp, (char *) bytesbuf, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_INT:
	    *bufp++ = PIPE_DRIVER_TERM_INT;
	    *(unsigned long *) bytesbuf = htonl(*ptr);
	    memcpy(bufp, (char *) bytesbuf, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_TUPLE:
	    *bufp++ = PIPE_DRIVER_TERM_TUPLE;
	    *bufp++ = *ptr;
	    ptr++;
	    break;
	case ERL_DRV_LIST:
	    *bufp++ = PIPE_DRIVER_TERM_LIST;
	    tmpi = ntohl(*ptr);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    ptr++;
	    break;
	case ERL_DRV_BINARY:
	    *bufp++ = PIPE_DRIVER_TERM_BINARY;
	    tmpi = ntohl(ptr[1]);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    ebin = (ErlDrvBinary *) ptr[0]; 
	    memcpy(bufp, ebin->orig_bytes + ptr[2], ptr[1]);
	    bufp += ptr[1];
	    ptr += 3;
	    break;
	case ERL_DRV_STRING:
	    /* The pointers pushed on the stack differ from binary & string! */
	    *bufp++ = PIPE_DRIVER_TERM_STRING;
	    tmpi = ntohl(ptr[1]);
	    memcpy(bufp, &tmpi, 4);
	    bufp += 4;
	    memcpy(bufp, (char *) ptr[0], ptr[1]);
	    bufp += ptr[1];
	    ptr += 2;
	    break;
	default:
	    edtk_debug("%s: ser: UNKNOWN tag = %d\r\n", __FUNCTION__, tag);
	    break;
	}
    }
    *(unsigned long *) bytesbuf = htonl(bytes);
    edtk_debug("%s: writing = %ld bytes\r\n", __FUNCTION__, bytes);
    if (write(1, bytesbuf, 4) != 4 || write(1, buf, bytes) != bytes) {
	edtk_debug("\r\n%s: write error, aborting!\r\n\r\n", __FUNCTION__);
	exit(1);
    }

    return 0;
}

long driver_async(ErlDrvPort ix,
                         unsigned int* key,
                         void (*async_invoke)(void*), 
                         void* async_data,
                         void (*async_free)(void*))
{
    edtk_debug("\r\n%s: This should never happen\r\n\r\n", __FUNCTION__);
    sleep(2);
    return 0;
}

/*
** Since we always pass atoms by name, just return the pointer to the name.
** It's not our problem to worry about the longevity of the pointer....
*/

ErlDrvTermData
driver_mk_atom(char *name)
{
    edtk_debug("%s: name = %s\r\n", __FUNCTION__, name);
    return (ErlDrvTermData) name;
}

/*
** Just return the port number as-is.
*/

ErlDrvTermData
driver_mk_port(ErlDrvPort port)
{
    edtk_debug("%s: port = 0x%lx\r\n", __FUNCTION__, port);
    return (ErlDrvTermData) port;
}


/*
** Nothing here for a pipe driver to do.
*/

int
driver_select(ErlDrvPort port, ErlDrvEvent event, int mode, int on)
{
    return 0;
}

#ifdef XXXCOMMENT

P = open_port({spawn, "./test-pipelib"}, [binary, {packet,4}]).
P = open_port({spawn, "../../pipe-main ./berkeley_db_drv-pipe.so"}, [binary, {packet,4}]).
port_command(P, <<200, 201, 202, 203>>).
receive M -> M after 0 -> foo end.
receive {_, {data, M1}} -> M1 after 0 -> foo end.


#endif
