%%%----------------------------------------------------------------------
%%% File    : berkeley_db_drv.erl
%%% Summary : Erlang driver for the berkeley_db
%%%
%%%
%%% NOTICE: This file was generated by the tools of the Erlang Driver
%%%         toolkit.  Do not edit this file by hand unless you know
%%%         what you're doing!
%%%
%%% Copyright (c) 2002, Scott Lystig Fritchie.  All rights reserved.
%%% See the file "LICENSE" at the top of the source distribution for
%%% full license terms.
%%%
%%%----------------------------------------------------------------------

-module(berkeley_db_drv).
-include("berkeley_db_drv.hrl").

%% Xref with erl_driver_tk.h's PIPE_DRIVER_TERM_* values
-define(T_NIL, 0).
-define(T_ATOM, 1).
-define(T_PORT, 2).
-define(T_INT, 3).
-define(T_TUPLE, 4).
-define(T_BINARY, 5).
-define(T_STRING, 6).
-define(T_LIST, 7).

%% External exports

-export([start/0, start/1, start_pipe/0, start_pipe/1]).
 
-export([shutdown/1]).
-export([debug/2]).
-export([
         null/1, 
         env_create/2, 
         env_close/2, 
         env_open/5, 
         env_remove/4, 
         env_set_data_dir/3, 
         env_set_errfile/4, 
         env_set_errpfx/3, 
         env_set_flags/4, 
         env_set_rpc_server/6, 
         env_set_shm_key/3, 
         env_set_tas_spins/3, 
         env_set_tmp_dir/3, 
         env_set_timeout/4, 
         env_set_verbose/4, 
         db_strerror/2, 
         db_version/1, 
         db_create/3, 
         db_close/2, 
         db_del/5, 
         db_fd/2, 
         db_get/6, 
         db_get_byteswapped/2, 
         db_get_type/2, 
         db_open/7, 
         db_put/6, 
         db_remove/5, 
         db_rename/6, 
         db_sync/3, 
         db_truncate/4, 
         db_upgrade/4, 
         db_verify/6, 
         db_set_cachesize/5, 
         db_set_errfile/4, 
         db_set_errpfx/3, 
         db_set_flags/3, 
         db_set_pagesize/3, 
         db_set_bt_minkey/3, 
         db_set_re_delim/3, 
         db_set_re_len/3, 
         db_set_re_pad/3, 
         db_set_re_source/3, 
         db_set_h_ffactor/3, 
         db_set_h_nelem/3, 
         db_set_lorder/3, 
         db_set_q_extentsize/3, 
         db_cursor/4, 
         c_close/2, 
         c_count/3, 
         c_del/3, 
         c_get/5, 
         c_dup/3, 
         c_put/5, 
         set_lk_conflicts/4, 
         set_lk_detect/3, 
         set_lk_max_lockers/3, 
         set_lk_max_locks/3, 
         set_lk_max_objects/3, 
         lock_detect/4, 
         lock_id/2, 
         lock_id_free/3, 
         lock_get/6, 
         lock_put/3, 
         txn_abort/2
        ]).


shutdown(Port) when port(Port) ->
    catch erlang:port_close(Port),
    %% I was under the impression you'd always get a message sent to
    %% you in this case, so this receive is to keep your mailbox from
    %% getting cluttered.  Hrm, well, sometimes the message does
    %% not arrive at all!
    receive
        {'EXIT', Port, normal} -> {ok, normal};
        {'EXIT', Port, Err}    -> {error, Err}
    after 0                    -> {ok, normall} % XXX is 0 too small?
        
    end.

debug(Port, Flags) when port(Port), integer(Flags) ->
    case catch erlang:port_command(Port, <<?_DEBUG, Flags:32>>) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX too drastic?
    end.

null(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_NULL>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_create(Port,
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_ENV_CREATE,
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_close(Port,
     Dbenv
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_CLOSE,
            DbenvIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_open(Port,
     Dbenv, 
     Db_Home, 
     Flags, 
     Mode
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {Db_HomeBinOrList, Db_HomeLen} = serialize_contiguously(Db_Home, 1),
    IOList_____ = [ <<?_ENV_OPEN,
            DbenvIndex:32/integer, 
            Db_HomeLen:32/integer>>,            % I/O list length
          Db_HomeBinOrList,
          <<
          Flags:32/unsigned-integer, 
          Mode:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_remove(Port,
     Dbenv, 
     Db_Home, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {Db_HomeBinOrList, Db_HomeLen} = serialize_contiguously(Db_Home, 1),
    IOList_____ = [ <<?_ENV_REMOVE,
            DbenvIndex:32/integer, 
            Db_HomeLen:32/integer>>,            % I/O list length
          Db_HomeBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_data_dir(Port,
     Dbenv, 
     Dir
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {DirBinOrList, DirLen} = serialize_contiguously(Dir, 1),
    IOList_____ = [ <<?_ENV_SET_DATA_DIR,
            DbenvIndex:32/integer, 
            DirLen:32/integer>>,                % I/O list length
          DirBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_errfile(Port,
     Dbenv, 
     File, 
     Fopen_Mode
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {Fopen_ModeBinOrList, Fopen_ModeLen} = serialize_contiguously(Fopen_Mode, 1),
    IOList_____ = [ <<?_ENV_SET_ERRFILE,
            DbenvIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            Fopen_ModeLen:32/integer>>,         % I/O list length
          Fopen_ModeBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_errpfx(Port,
     Dbenv, 
     Errpfx
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {ErrpfxBinOrList, ErrpfxLen} = serialize_contiguously(Errpfx, 1),
    IOList_____ = [ <<?_ENV_SET_ERRPFX,
            DbenvIndex:32/integer, 
            ErrpfxLen:32/integer>>,             % I/O list length
          ErrpfxBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_flags(Port,
     Dbenv, 
     Flags, 
     Onoff
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_SET_FLAGS,
            DbenvIndex:32/integer, 
          Flags:32/unsigned-integer, 
          Onoff:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_rpc_server(Port,
     Dbenv, 
     Host, 
     Cl_Timeout, 
     Sv_Timeout, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {HostBinOrList, HostLen} = serialize_contiguously(Host, 1),
    IOList_____ = [ <<?_ENV_SET_RPC_SERVER,
            DbenvIndex:32/integer, 
            HostLen:32/integer>>,               % I/O list length
          HostBinOrList,
          <<
          Cl_Timeout:32/integer, 
          Sv_Timeout:32/integer, 
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_shm_key(Port,
     Dbenv, 
     Shm_Key
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_SET_SHM_KEY,
            DbenvIndex:32/integer, 
          Shm_Key:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_tas_spins(Port,
     Dbenv, 
     Tas_Spins
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_SET_TAS_SPINS,
            DbenvIndex:32/integer, 
          Tas_Spins:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_tmp_dir(Port,
     Dbenv, 
     Dir
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {DirBinOrList, DirLen} = serialize_contiguously(Dir, 1),
    IOList_____ = [ <<?_ENV_SET_TMP_DIR,
            DbenvIndex:32/integer, 
            DirLen:32/integer>>,                % I/O list length
          DirBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_timeout(Port,
     Dbenv, 
     Timeout, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_SET_TIMEOUT,
            DbenvIndex:32/integer, 
          Timeout:32/unsigned-integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

env_set_verbose(Port,
     Dbenv, 
     Flags, 
     Onoff
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_ENV_SET_VERBOSE,
            DbenvIndex:32/integer, 
          Flags:32/unsigned-integer, 
          Onoff:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_strerror(Port,
     Error
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_DB_STRERROR,
          Error:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_version(Port
        ) when port(Port) -> % TODO: Add additional constraints here
    IOList_____ = <<?_DB_VERSION>>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_create(Port,
     Dbenv, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_DB_CREATE,
            DbenvIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_close(Port,
     Db
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_CLOSE,
            DbIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_del(Port,
     Db, 
     Txnid, 
     Ekey_In, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    TxnidIndex = case Txnid of
                {valmap_db_txn, N} -> N;
                void                        -> -1;
                Else                        -> throw({bogus_valmap, Else})
    end,
    {Ekey_InBinOrList, Ekey_InLen} = serialize_contiguously(Ekey_In, 0),
    IOList_____ = [ <<?_DB_DEL,
            DbIndex:32/integer, 
            TxnidIndex:32/integer, 
            Ekey_InLen:32/integer>>,            % I/O list length
          Ekey_InBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_fd(Port,
     Db
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_FD,
            DbIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_get(Port,
     Db, 
     Txnid, 
     Ekey_In, 
     Edata_In, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    TxnidIndex = case Txnid of
                {valmap_db_txn, N} -> N;
                void                        -> -1;
                Else                        -> throw({bogus_valmap, Else})
    end,
    {Ekey_InBinOrList, Ekey_InLen} = serialize_contiguously(Ekey_In, 0),
    {Edata_InBinOrList, Edata_InLen} = serialize_contiguously(Edata_In, 0),
    IOList_____ = [ <<?_DB_GET,
            DbIndex:32/integer, 
            TxnidIndex:32/integer, 
            Ekey_InLen:32/integer>>,            % I/O list length
          Ekey_InBinOrList,
          <<
            Edata_InLen:32/integer>>,           % I/O list length
          Edata_InBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_get_byteswapped(Port,
     Db
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_GET_BYTESWAPPED,
            DbIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_get_type(Port,
     Db
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_GET_TYPE,
            DbIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_open(Port,
     Db, 
     File, 
     Database, 
     Type, 
     Flags, 
     Mode
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {DatabaseBinOrList, DatabaseLen} = serialize_contiguously(Database, 1),
    IOList_____ = [ <<?_DB_OPEN,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            DatabaseLen:32/integer>>,           % I/O list length
          DatabaseBinOrList,
          <<
          Type:32/integer, 
          Flags:32/unsigned-integer, 
          Mode:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_put(Port,
     Db, 
     Txnid, 
     Ekey_In, 
     Edata_In, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    TxnidIndex = case Txnid of
                {valmap_db_txn, N} -> N;
                void                        -> -1;
                Else                        -> throw({bogus_valmap, Else})
    end,
    {Ekey_InBinOrList, Ekey_InLen} = serialize_contiguously(Ekey_In, 0),
    {Edata_InBinOrList, Edata_InLen} = serialize_contiguously(Edata_In, 0),
    IOList_____ = [ <<?_DB_PUT,
            DbIndex:32/integer, 
            TxnidIndex:32/integer, 
            Ekey_InLen:32/integer>>,            % I/O list length
          Ekey_InBinOrList,
          <<
            Edata_InLen:32/integer>>,           % I/O list length
          Edata_InBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_remove(Port,
     Db, 
     File, 
     Database, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {DatabaseBinOrList, DatabaseLen} = serialize_contiguously(Database, 1),
    IOList_____ = [ <<?_DB_REMOVE,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            DatabaseLen:32/integer>>,           % I/O list length
          DatabaseBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_rename(Port,
     Db, 
     File, 
     Database, 
     Newname, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {DatabaseBinOrList, DatabaseLen} = serialize_contiguously(Database, 1),
    {NewnameBinOrList, NewnameLen} = serialize_contiguously(Newname, 1),
    IOList_____ = [ <<?_DB_RENAME,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            DatabaseLen:32/integer>>,           % I/O list length
          DatabaseBinOrList,
          <<
            NewnameLen:32/integer>>,            % I/O list length
          NewnameBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_sync(Port,
     Db, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SYNC,
            DbIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_truncate(Port,
     Db, 
     Txnid, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    TxnidIndex = case Txnid of
                {valmap_db_txn, N} -> N;
                void                        -> -1;
                Else                        -> throw({bogus_valmap, Else})
    end,
    IOList_____ = <<?_DB_TRUNCATE,
            DbIndex:32/integer, 
            TxnidIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_upgrade(Port,
     Db, 
     File, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    IOList_____ = [ <<?_DB_UPGRADE,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_verify(Port,
     Db, 
     File, 
     Database, 
     Outfile, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {DatabaseBinOrList, DatabaseLen} = serialize_contiguously(Database, 1),
    {OutfileBinOrList, OutfileLen} = serialize_contiguously(Outfile, 1),
    IOList_____ = [ <<?_DB_VERIFY,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            DatabaseLen:32/integer>>,           % I/O list length
          DatabaseBinOrList,
          <<
            OutfileLen:32/integer>>,            % I/O list length
          OutfileBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_cachesize(Port,
     Db, 
     Gbytes, 
     Bytes, 
     Ncache
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_CACHESIZE,
            DbIndex:32/integer, 
          Gbytes:32/unsigned-integer, 
          Bytes:32/unsigned-integer, 
          Ncache:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_errfile(Port,
     Db, 
     File, 
     Fopen_Mode
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {FileBinOrList, FileLen} = serialize_contiguously(File, 1),
    {Fopen_ModeBinOrList, Fopen_ModeLen} = serialize_contiguously(Fopen_Mode, 1),
    IOList_____ = [ <<?_DB_SET_ERRFILE,
            DbIndex:32/integer, 
            FileLen:32/integer>>,               % I/O list length
          FileBinOrList,
          <<
            Fopen_ModeLen:32/integer>>,         % I/O list length
          Fopen_ModeBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_errpfx(Port,
     Db, 
     Errpfx
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {ErrpfxBinOrList, ErrpfxLen} = serialize_contiguously(Errpfx, 1),
    IOList_____ = [ <<?_DB_SET_ERRPFX,
            DbIndex:32/integer, 
            ErrpfxLen:32/integer>>,             % I/O list length
          ErrpfxBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_flags(Port,
     Db, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_FLAGS,
            DbIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_pagesize(Port,
     Db, 
     Pagesize
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_PAGESIZE,
            DbIndex:32/integer, 
          Pagesize:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_bt_minkey(Port,
     Db, 
     Bt_Minkey
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_BT_MINKEY,
            DbIndex:32/integer, 
          Bt_Minkey:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_re_delim(Port,
     Db, 
     Re_Delim
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_RE_DELIM,
            DbIndex:32/integer, 
          Re_Delim:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_re_len(Port,
     Db, 
     Re_Len
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_RE_LEN,
            DbIndex:32/integer, 
          Re_Len:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_re_pad(Port,
     Db, 
     Re_Pad
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_RE_PAD,
            DbIndex:32/integer, 
          Re_Pad:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_re_source(Port,
     Db, 
     Re_Source
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    {Re_SourceBinOrList, Re_SourceLen} = serialize_contiguously(Re_Source, 1),
    IOList_____ = [ <<?_DB_SET_RE_SOURCE,
            DbIndex:32/integer, 
            Re_SourceLen:32/integer>>,          % I/O list length
          Re_SourceBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_h_ffactor(Port,
     Db, 
     H_Ffactor
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_H_FFACTOR,
            DbIndex:32/integer, 
          H_Ffactor:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_h_nelem(Port,
     Db, 
     H_Nelem
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_H_NELEM,
            DbIndex:32/integer, 
          H_Nelem:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_lorder(Port,
     Db, 
     Lorder
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_LORDER,
            DbIndex:32/integer, 
          Lorder:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_set_q_extentsize(Port,
     Db, 
     Extentsize
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    IOList_____ = <<?_DB_SET_Q_EXTENTSIZE,
            DbIndex:32/integer, 
          Extentsize:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

db_cursor(Port,
     Db, 
     Txnid, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db, DbIndex} = Db,
    TxnidIndex = case Txnid of
                {valmap_db_txn, N} -> N;
                void                        -> -1;
                Else                        -> throw({bogus_valmap, Else})
    end,
    IOList_____ = <<?_DB_CURSOR,
            DbIndex:32/integer, 
            TxnidIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_close(Port,
     Dbc
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    IOList_____ = <<?_C_CLOSE,
            DbcIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_count(Port,
     Dbc, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    IOList_____ = <<?_C_COUNT,
            DbcIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_del(Port,
     Dbc, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    IOList_____ = <<?_C_DEL,
            DbcIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_get(Port,
     Dbc, 
     Ekey_In, 
     Edata_In, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    {Ekey_InBinOrList, Ekey_InLen} = serialize_contiguously(Ekey_In, 0),
    {Edata_InBinOrList, Edata_InLen} = serialize_contiguously(Edata_In, 0),
    IOList_____ = [ <<?_C_GET,
            DbcIndex:32/integer, 
            Ekey_InLen:32/integer>>,            % I/O list length
          Ekey_InBinOrList,
          <<
            Edata_InLen:32/integer>>,           % I/O list length
          Edata_InBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_dup(Port,
     Dbc, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    IOList_____ = <<?_C_DUP,
            DbcIndex:32/integer, 
          Flags:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

c_put(Port,
     Dbc, 
     Ekey_In, 
     Edata_In, 
     Flags
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_cursor, DbcIndex} = Dbc,
    {Ekey_InBinOrList, Ekey_InLen} = serialize_contiguously(Ekey_In, 0),
    {Edata_InBinOrList, Edata_InLen} = serialize_contiguously(Edata_In, 0),
    IOList_____ = [ <<?_C_PUT,
            DbcIndex:32/integer, 
            Ekey_InLen:32/integer>>,            % I/O list length
          Ekey_InBinOrList,
          <<
            Edata_InLen:32/integer>>,           % I/O list length
          Edata_InBinOrList,
          <<
          Flags:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

set_lk_conflicts(Port,
     Dbenv, 
     Conflicts, 
     Nmodes
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {ConflictsBinOrList, ConflictsLen} = serialize_contiguously(Conflicts, 0),
    IOList_____ = [ <<?_SET_LK_CONFLICTS,
            DbenvIndex:32/integer, 
            ConflictsLen:32/integer>>,          % I/O list length
          ConflictsBinOrList,
          <<
          Nmodes:32/integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

set_lk_detect(Port,
     Dbenv, 
     Detect
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_SET_LK_DETECT,
            DbenvIndex:32/integer, 
          Detect:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

set_lk_max_lockers(Port,
     Dbenv, 
     Max
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_SET_LK_MAX_LOCKERS,
            DbenvIndex:32/integer, 
          Max:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

set_lk_max_locks(Port,
     Dbenv, 
     Max
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_SET_LK_MAX_LOCKS,
            DbenvIndex:32/integer, 
          Max:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

set_lk_max_objects(Port,
     Dbenv, 
     Max
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_SET_LK_MAX_OBJECTS,
            DbenvIndex:32/integer, 
          Max:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

lock_detect(Port,
     Dbenv, 
     Flags, 
     Atype
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_LOCK_DETECT,
            DbenvIndex:32/integer, 
          Flags:32/unsigned-integer, 
          Atype:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

lock_id(Port,
     Dbenv
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_LOCK_ID,
            DbenvIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

lock_id_free(Port,
     Dbenv, 
     Id
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    IOList_____ = <<?_LOCK_ID_FREE,
            DbenvIndex:32/integer, 
          Id:32/unsigned-integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

lock_get(Port,
     Dbenv, 
     Locker, 
     Flags, 
     Eobj, 
     Lock_Mode
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {EobjBinOrList, EobjLen} = serialize_contiguously(Eobj, 0),
    IOList_____ = [ <<?_LOCK_GET,
            DbenvIndex:32/integer, 
          Locker:32/unsigned-integer, 
          Flags:32/unsigned-integer, 
            EobjLen:32/integer>>,               % I/O list length
          EobjBinOrList,
          <<
          Lock_Mode:32/unsigned-integer
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

lock_put(Port,
     Dbenv, 
     Elock
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_dbenv, DbenvIndex} = Dbenv,
    {ElockBinOrList, ElockLen} = serialize_contiguously(Elock, 0),
    IOList_____ = [ <<?_LOCK_PUT,
            DbenvIndex:32/integer, 
            ElockLen:32/integer>>,              % I/O list length
          ElockBinOrList,
          <<
        >> ],
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

txn_abort(Port,
     Db_Txn
        ) when port(Port) -> % TODO: Add additional constraints here
    {valmap_db_txn, Db_TxnIndex} = Db_Txn,
    IOList_____ = <<?_TXN_ABORT,
            Db_TxnIndex:32/integer
        >>,
    case catch erlang:port_command(Port, IOList_____) of
        true -> get_port_reply(Port);
        Err  -> throw(Err)              % XXX Is this too drastic?
    end.

    
%%%
%%% Internal functions.
%%%

load_path(File) ->
    case lists:filter(fun(D) ->
                              case file:read_file_info(D ++ "/" ++ File) of
                                  {ok, _} -> true;
                                  _ -> false
                              end
                      end, code:get_path()) of
        [Dir|_] ->
            {ok, Dir};
        [] ->
            io:format("Error: ~s not found in code path\n", [File]),
            {error, enoent}
    end.

%%%
%%% Note that an 'xtra_return' that only returns one item in its
%%% tuple will return {Port, ok, {Thingie}}, so we'll return
%%% {ok, {Thingie}}, which is *sooooooo* maddening because I keep
%%% forgetting the extra tuple wrapper.  So, if there's only one
%%% thingie in the return tuple, strip it off: {ok, Thingie}
%%%

get_port_reply(Port) when port(Port) ->
    receive
        {Port, ok} = T -> proc_reply(T);
        {Port, ok, {M}} = T -> proc_reply(T);
        {Port, ok, M} = T -> proc_reply(T);
        {Port, error, {Reason}} = T -> proc_reply(T);
        {Port, error, Reason} = T -> proc_reply(T);
        %% Pipe driver messages
        {Port, {data, Bytes}} -> proc_reply(pipedrv_deser(Port, Bytes));
        {'EXIT', Port, Reason} -> throw({port_error, Reason});  % XXX too drastic?
        {Port, Reason} -> throw({port_error, Reason})   % XXX too drastic?
    end.

%% This function exists to provide consistency of replies 
%% given by linked-in and pipe drivers.  The "receive" statement
%% in get_port_reply/1 is specific because we want it to be
%% very selective about what it will grab out of the mailbox.
proc_reply({Port, ok}) when port(Port) ->
    ok;
proc_reply({Port, ok, {M}}) when port(Port) ->
    {ok, M};
proc_reply({Port, ok, M}) when port(Port) ->
    {ok, M};
proc_reply({Port, error, {Reason}}) when port(Port) ->
    {error, Reason};
proc_reply({Port, error, Reason}) when port(Port) ->
    {error, Reason}.


%%% io_list_len() is an extremely useful function.  BEAM has got this
%%% implemented quite efficiently in C.  It would be *fabulous* to be able
%%% to use it from Erlang via a BIF.

io_list_len(B) when binary(B) -> {B, size(B)};
io_list_len(L) when list(L) -> io_list_len(L, 0).
io_list_len([H|T], N) ->
    if
        H >= 0, H =< 255 -> io_list_len(T, N+1);
        list(H) -> io_list_len(T, io_list_len(H,N));
        binary(H) -> io_list_len(T, size(H) + N);
        true -> throw({error, partial_len, N})
    end;
io_list_len(H, N) when binary(H) -> 
    size(H) + N;
io_list_len([], N) -> 
    N.

%%% We need to make the binary thing we're passing in contiguous
%%% because the C function we're calling is expecting a single
%%% contiguous buffer.  If IOList is ["Hello, ", <<"World">>, "!"],
%%% that binary in the middle element will end up with the argument
%%% spanning three parts of an ErlIOVec.  If that happens, then we'd
%%% have to have the driver do the dirty work of putting the argument
%%% into a single contiguous buffer.
%%%
%%% Frankly, we're lazy, and this code is short and won't be much
%%% slower than doing it in C.

%%% 2nd arg: if 1, NUL-terminate the IOList

serialize_contiguously(B, 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously([B], 0) when binary(B) ->
    {B, size(B)};
serialize_contiguously(IOList, 1) ->
    serialize_contiguously([IOList, 0], 0);
serialize_contiguously(IOList, 0) ->
    B = list_to_binary(IOList),
    {B, size(B)}.


%% pipedrv_deser/2 -- Deserialize the term that the pipe driver is
%% is returning to Erlang.  The pipe driver doesn't know it's a pipe
%% driver, it thinks it's a linked-in driver, so it tries to return
%% an arbitrary Erlang term to us.  The pipe-main program is sneaky:
%% it has a driver_output_term() function that serializes the term
%% that the driver built.  With the help of a list-as-stack, we
%% deserialize that term.

pipedrv_deser(Port, B) ->
    pipedrv_deser(Port, B, []).

pipedrv_deser(Port, <<>>, []) ->
    throw(icky_i_think);
pipedrv_deser(Port, <<>>, [T]) ->
    T;
pipedrv_deser(Port, <<?T_NIL:8, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [foo___foo_nil___|Stack]);
pipedrv_deser(Port, <<?T_ATOM:8, Len:8, Rest/binary>>, Stack) ->
    <<A:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [list_to_atom(binary_to_list(A))|Stack]);
pipedrv_deser(Port, <<?T_PORT:8, P:32/unsigned, Rest/binary>>, Stack) ->
    %% The pipe driver tried sending us a port, but it cannot know what
    %% port ID was assigned to this port, so we'll assume it is Port.
    pipedrv_deser(Port, Rest, [Port|Stack]);
pipedrv_deser(Port, <<?T_INT:8, I:32/signed, Rest/binary>>, Stack) ->
    pipedrv_deser(Port, Rest, [I|Stack]);
pipedrv_deser(Port, <<?T_TUPLE:8, N:8, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [list_to_tuple(L)|NewStack]);
pipedrv_deser(Port, <<?T_LIST:8, N:32, Rest/binary>>, Stack) ->
    {L, NewStack} = popN(N, Stack),
    pipedrv_deser(Port, Rest, [L|NewStack]);
pipedrv_deser(Port, <<?T_BINARY:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [Bin|Stack]);
pipedrv_deser(Port, <<?T_STRING:8, Len:32/signed, Rest/binary>>, Stack) ->
    <<Bin:Len/binary, Rest2/binary>> = Rest,
    pipedrv_deser(Port, Rest2, [binary_to_list(Bin)|Stack]);
pipedrv_deser(Port, X, Y) ->
    throw({bah, X, Y}).

popN(N, Stack) ->
    popN(N, Stack, []).
popN(0, Stack, Acc) ->
    {Acc, Stack};
popN(N, [foo___foo_nil___|T], Acc) ->
    %% This is the nonsense we put on the stack to represent NIL.  Ignore it.
    popN(N - 1, T, Acc);
popN(N, [H|T], Acc) ->
    popN(N - 1, T, [H|Acc]).


%%%
%%% Begin code included via <custom_erl> tags
%%%


start() ->
    start(0, false).

start(EnvCreateFlags) ->
    start(EnvCreateFlags, false).

start_pipe() ->
    start_pipe(0, false).

start_pipe(EnvCreateFlags) ->
    start_pipe(EnvCreateFlags, false).

start(EnvCreateFlags, NoCreateEnv) when integer(EnvCreateFlags) ->
    {ok, Path} = load_path(?DRV_NAME ++ ".so"),
    erl_ddll:start(),
    ok = erl_ddll:load_driver(Path, ?DRV_NAME),
    case open_port({spawn, ?DRV_NAME}, []) of
        P when port(P) ->
            case NoCreateEnv of
                true ->
                    {ok, P};
                _ ->
                    %% The driver will *not* work correctly without calling
                    %% env_create(), so we'll force it to happen here.
                    {ok, DBENV} = env_create(P, EnvCreateFlags),
                    {ok, P, DBENV}
            end;
        Err ->
            Err
    end.

start_pipe(EnvCreateFlags, NoCreateEnv) when integer(EnvCreateFlags) ->
    {ok, PipeMain} = load_path("pipe-main"),
    {ok, ShLib} = load_path(?DRV_NAME ++ ".so"),
    Cmd = PipeMain ++ "/pipe-main " ++ ShLib ++ "/" ++ ?DRV_NAME ++ ".so",
    case open_port({spawn, Cmd}, [exit_status, binary, use_stdio, {packet, 4}]) of
        P when port(P) ->
            case NoCreateEnv of
                true ->
                    {ok, P};
                _ ->
                    %% The driver will *not* work correctly without calling
                    %% env_create(), so we'll force it to happen here.
                    {ok, DBENV} = env_create(P, EnvCreateFlags),
                    {ok, P, DBENV}
            end;
        Err ->
            Err
    end.
 

