
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>

#include <db.h>
#include <rpc/rpc.h>

#include <erl_driver.h>
#include <erl_driver_tk.h>
#include <berkeley_db_drv.h>
#include <my-berkeley_db.h>

#ifdef	DRIVER_USING_PTHREADS
#include <pthread.h>
#endif	/* DRIVER_USING_PTHREADS */

void
null(void)
{
    printf("Hello, world!  This is null()\r\n");
}

/*
** We flip around the return value of db_env_create so that we normally
** return the new DB_ENV *, which makes EDTK's simple valmap mechanism
** happy.
*/

DB_ENV *
my_db_env_create(u_int32_t flags, int *my_error_value)
{
    DB_ENV	*dbenv;
    int		err;

    if ((err = db_env_create(&dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return dbenv;
}

/*
** QQQ Fortunately, the flags arg to env close is always constant,
** so we don't have to make any changes to EDTK.  Laziness rules!
*/

int
my_db_env_close(DB_ENV *dbenv)
{
    edtk_debug("%s: dbenv = 0x%lx", __FUNCTION__, dbenv);
    return dbenv->close(dbenv, 0);
}

void 
my_debug_errcall(const char *errpfx, char *msg)
{
    fprintf(stderr, "\r\n\r\n%s: %s: %s\r\n\r\n", __FUNCTION__, (errpfx == NULL) ? "(NULL)" : errpfx, msg);
}

int
my_env_set_errfile(DB_ENV *dbenv, char *file, char *mode)
{
    FILE	*fp;

    if ((fp = fopen(file, mode)) == NULL) {
	return errno;
    }
    dbenv->set_errfile(dbenv, fp);
    return 0;
}

int
my_db_set_errfile(DB *db, char *file, char *mode)
{
    FILE	*fp;

    if ((fp = fopen(file, mode)) == NULL) {
	return errno;
    }
    db->set_errfile(db, fp);
    return 0;
}

DB *
my_db_create(DB_ENV *dbenv, u_int32_t flags, int *my_error_value)
{
    DB	*db;
    int	err;

    if ((err = db_create(&db, dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return db;
}

/*
** QQQ We won't give a "flags" option to our users.  Perhaps that will
**     need changing someday?
*/

int
my_db_close(DB *db)
{
    edtk_debug("%s: db = 0x%lx", __FUNCTION__, db);
    return db->close(db, 0);
}

DBC *
my_cursor_create(DB *db, DB_TXN *txnid, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = db->cursor(db, txnid, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return cursorp;
}

DBC *
my_cursor_dup(DBC *cursor, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = cursor->c_dup(cursor, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return cursorp;
}

int
my_cursor_close(DBC *cursor)
{
    edtk_debug("%s: cursor = 0x%lx", __FUNCTION__, cursor);
    return cursor->c_close(cursor);
}

int
my_txn_abort(DB_TXN *db_txn)
{
    edtk_debug("%s: db_txn = 0x%lx", __FUNCTION__, db_txn);
    return db_txn->abort(db_txn);
}

int
my_db_verify(DB *db, char *file, char *database, char *outfilename,
	     u_int32_t flags)
{
    FILE	*outfile = NULL;

    if (outfilename != NULL && (outfile = fopen(outfilename, "w")) == NULL) {
	return errno;
    }
    return db->verify(db, file, database, outfile, flags);
}

int
make_dbt_key(struct descriptor *desc, callstate_t *c, ErlDrvTermData *msg,
	     int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.key_in, c->i.ekey_in);
}

int
make_dbt_data(struct descriptor *desc, callstate_t *c, ErlDrvTermData *msg,
	     int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.data_in, c->i.edata_in);
}

static int
make_dbt(struct descriptor *desc, callstate_t *c, ErlDrvTermData *msg,
	 int *members, int *msgcount, DBT *dbt, char *orig_buf)
{
    char		*new;
    ErlDrvBinary	*tmpbin = NULL;

    if (dbt->data == orig_buf) {
	if ((new = edtk_driver_alloc_wrapper(dbt->size)) == NULL) {
	    *members = 0;
	    return 0;
	}
	memcpy(new, dbt->data, dbt->size);
	tmpbin = edtk_alloced_ptr2ErlDrvBinary(new);
    } else {
	tmpbin = edtk_alloced_ptr2ErlDrvBinary(dbt->data);
    }
    *msgcount = LOAD_BINARY(msg, *msgcount, tmpbin, 0, dbt->size);
    *members = 1;
    return 1;
}
