#!/bin/sh

case $0 in
    .*/*)
	TMP=`dirname $0`
        STARTDIR=`cd $TMP ; pwd`
    ;;
    */*)
        STARTDIR=`dirname $0`
    ;;
    *)
        MYSELF=`which $0`
        STARTDIR=`dirname $MYSELF`
    ;;
esac
UP1=`dirname $STARTDIR`
UP2=`dirname $UP1`
MYDIR=$UP2

INTERACTIVE="-noinput"
NODENAME="cd"
THREADS=5
BEAM=""
BOOTFILE="$MYDIR/ethbridge/src/ethbridge"		# location for development only
ERLANG_ROOT=""
# XXX The following should be put in /var somewhere.
while [ ! -z "$1" ]; do
	case "$1" in
	    -B)
		BOOTFILE="$2"
		shift
	    ;;
	    -b)
		BEAM="$2"
		if [ ! -f $BEAM ]; then
			echo "BEAM interpreter $BEAM does not exist"
			exit 1
		fi
		shift
	    ;;
	    -i)
		INTERACTIVE=""
	    ;;
	    -n)
		NODENAME="$2"
		shift
	    ;;
	    -P)
	        # For use in a "P"roduction environment, to save typing
	        # by a sysadmin.
	        BEAM="/usr/local/sendmail/bams/lib/erlang/erts-4.9.1/bin/beam"
		BOOTFILE="/usr/local/sendmail/bams/lib/erlang/releases/cd-0.01/cd"
	    ;;
	    -r)
		ERLANG_ROOT="$2"
		shift
	    ;;
	    -t)
		THREADS="$2"
		shift
	    ;;
	    -*)
		echo "usage: $0 [-i] [-f fstab_file] [-n erlang-node-name]"
		echo "	[-b beam-interpreter [-r erlang-root-dir]]"
		echo "	[-n erlang-node-name] [-p CP-port-number]"
		exit 1
	esac
	shift
done

if [ ! -z "$BEAM" -a -z "$ERLANG_ROOT" ]; then
	# Assume that the Erlang root dir is 2 levels up from the BEAM binary
	ERLANG_ROOT=`dirname $BEAM`/../..
fi

DIRLIST=""
for DIR in ethbridge/ebin libnet libpcap
do
	DIRLIST="$DIRLIST -pz $MYDIR/$DIR"
done

if [ -z "$BOOTFILE" ]; then
	BOOTFILE="$MYDIR/ethbridge/src/ethbridge"
fi

if [ -z "$BEAM" ]; then
	exec erl +A$THREADS \
		-sname $NODENAME -boot $BOOTFILE \
		$DIRLIST $INTERACTIVE
else
	EPMD=`dirname $BEAM`/epmd
	$EPMD -daemon
	exec $BEAM -A$THREADS \
		-- -root $ERLANG_ROOT -progname erl -- -home $HOME \
		-sname $NODENAME -boot $BOOTFILE \
		$DIRLIST $INTERACTIVE
fi
