
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include <libnet.h>
#include <my-libnet.h>
#include <erl_driver_tk.h>

static int example3_main(int argc, char **argv, int fd);

void
null(void)
{
    printf("Hello, world!  This is null()\r\n");
}

int
test_send_ping(char *src, char *dst, int raw_sock)
{
    char	*argv[16];
    int		argc = 1;

    argv[argc - 1] = "test_send_ping";
    argc++;
    argv[argc - 1] = "-s";
    argc++;
    argv[argc - 1] = src;
    argc++;
    argv[argc - 1] = "-d";
    argc++;
    argv[argc - 1] = dst;
    argc++;
    argv[argc] = NULL;
#if	!defined(linux) && !defined(SOLARIS)
    optreset = 1;
#endif	/* !linux && !SOLARIS */
    optind = 1;
    return example3_main(argc, argv, raw_sock);
}


void usage(char *);

static int
example3_main(int argc, char **argv, int fd)
{
    int network = fd, n, c, number_of_packets, packet_size;
    struct libnet_arena arena, *arena_p;
    u_char *packets[10];
    u_long src_ip, dst_ip;
    
    printf("libnet example code:\tmodule 3\n\n");
    printf("packet injection interface:\tlink layer\n");
    printf("packet type:\t\t\tICMP_ECHO [no payload] using an arena\n");

    src_ip = 0;
    dst_ip = 0;
    while((c = getopt(argc, argv, "d:s:")) != EOF)
    {
fprintf(stderr, "optarg = %s\r\n", optarg);
        switch (c)
        {
            case 'd':
                if (!(dst_ip = libnet_name_resolve((u_char *) optarg, LIBNET_RESOLVE)))
                {
                    libnet_error(LIBNET_ERR_FATAL, "Bad destination IP address: %s\n", optarg);
                }
                break;
            case 's':
                if (!(src_ip = libnet_name_resolve((u_char *) optarg, LIBNET_RESOLVE)))
                {
                    libnet_error(LIBNET_ERR_FATAL, "Bad source IP address: %s\n", optarg);
                }
                break;
        }
    }
    if (!src_ip || !dst_ip)
    {
        usage(argv[0]);
	fprintf(stderr, "argv0 = %s\r\n", argv[0]);
	fprintf(stderr, "argv1 = %s\r\n", argv[1]);
	fprintf(stderr, "argv2 = %s\r\n", argv[2]);
	fprintf(stderr, "argv3 = %s\r\n", argv[3]);
	fprintf(stderr, "argv4 = %s\r\n", argv[4]);
        return(EXIT_FAILURE);
    }

    /*
     *  We're just going to build an ICMP packet with no payload using the
     *  raw sockets API, so we only need memory for a ICMP header and an IP
     *  header.
     */
    packet_size = LIBNET_IP_H + LIBNET_ICMP_ECHO_H;

    /*
     *  Let's just build say, 10 packets.
     */
    number_of_packets = 10;

    arena_p = &arena;
    if (libnet_init_packet_arena(&arena_p, number_of_packets, packet_size) == -1)
    {
        libnet_error(LIBNET_ERR_FATAL, "libnet_init_packet_arena failed\n");
    }
    else
    {
        printf("Allocated an arena of %ld bytes..\n", LIBNET_GET_ARENA_SIZE(arena));
    }

    for (n = 0; n < number_of_packets; n++)
    {
        printf("%ld bytes remaining in arena\n", LIBNET_GET_ARENA_REMAINING_BYTES(arena));
        packets[n] = libnet_next_packet_from_arena(&arena_p, packet_size);
        if (!packets[n])
        {
            libnet_error(LIBNET_ERR_WARNING, "Arena is empty\n");
            continue;
        }

        libnet_build_ip(ICMP_ECHO_H,          	    /* Size of the payload */
                IPTOS_LOWDELAY | IPTOS_THROUGHPUT,  /* IP tos */
                242,                                /* IP ID */
                0,                                  /* frag stuff */
                48,                                 /* TTL */
                IPPROTO_ICMP,                       /* transport protocol */
                src_ip,                             /* source IP */
                dst_ip,                             /* destination IP */
                NULL,                               /* pointer to payload */
                0,                                  /* payload length */
                packets[n]);                        /* packet header memory */

        libnet_build_icmp_echo(ICMP_ECHO,           /* type */
                0,                                  /* code */
                242,                                /* id */
                5,                                  /* seq */
                NULL,	                            /* pointer to payload */
                0,                                  /* payload length */
                packets[n] + LIBNET_IP_H);          /* packet header memory */

        if (libnet_do_checksum(packets[n], IPPROTO_ICMP, LIBNET_ICMP_ECHO_H) == -1)
        {
            libnet_error(LIBNET_ERR_FATAL, "libnet_do_checksum failed\n");
        }

        c = libnet_write_ip(network, packets[n], packet_size);
        if (c < packet_size)
        {
            libnet_error(LN_ERR_WARNING, "libnet_write_ip only wrote %d bytes\n", c);
        }
        else
        {
            printf("construction and injection of packet %d of %d completed, wrote all %d bytes\n",  n + 1, number_of_packets, c);
        }
    }

    libnet_destroy_packet_arena(&arena_p);
    return (c == -1 ? EXIT_FAILURE : EXIT_SUCCESS);
}


void
usage(char *name)
{
    fprintf(stderr, "usage: %s -s source_ip -d destination_ip\n ", name);
}

int
my_libnet_select_device_any(struct sockaddr_in *sin, char **device,
			    char *ebuf)
{
    return libnet_select_device(sin, device, ebuf);
}

u_char *
test_make_icmp_echo_no_checksum(char *src, char *dst, int *pkt_len)
{
    int packet_size;
    u_char *packet;
    u_long src_ip, dst_ip;
    
    src_ip = 0;
    dst_ip = 0;
    if (!(dst_ip = libnet_name_resolve((u_char *) dst, LIBNET_RESOLVE)) ||
	!(src_ip = libnet_name_resolve((u_char *) src, LIBNET_RESOLVE)) ||
	(packet = edtk_driver_alloc_wrapper(1500 /* Way too big */)) == NULL) {
	return NULL;
    }
    packet_size = LIBNET_IP_H + LIBNET_ICMP_ECHO_H;

    libnet_build_ip(ICMP_ECHO_H,          	    /* Size of the payload */
		    IPTOS_LOWDELAY | IPTOS_THROUGHPUT,  /* IP tos */
		    242,                            /* IP ID */
		    0,                              /* frag stuff */
		    48,                             /* TTL */
		    IPPROTO_ICMP,                   /* transport protocol */
		    src_ip,                         /* source IP */
		    dst_ip,                         /* destination IP */
		    NULL,                           /* pointer to payload */
		    0,                              /* payload length */
		    packet);                        /* packet header memory */

    libnet_build_icmp_echo(ICMP_ECHO,               /* type */
			   0,                       /* code */
			   242,                     /* id */
			   5,                       /* seq */
			   NULL,	            /* pointer to payload */
			   0,                       /* payload length */
			   packet + LIBNET_IP_H);   /* packet header memory */
    *pkt_len = packet_size;
    return packet;
}
