-module('libpcap_test').
-define(DRV, libpcap_drv).

-export([regression/0]).

%% Change these interface names to suit your local environment
-define(MY_OS, "FreeBSD").
-define(IFACE_EXIST, "xl0").
-define(IFACE_NOTEXIST, "blar4").
% Path to the "test-captures" directory
-define(TEST_CAPTURES_DIR,
	"/user/fritchie/src/e-d/edtk/examples/libpcap/test-captures").
%-define(MY_OS, "Linux").
%-define(IFACE_EXIST, "eth0").
%-define(IFACE_NOTEXIST, "blar4").
%% Path to the "test-captures" directory
%-define(TEST_CAPTURES_DIR,
%      "/tmp/edtk-1.0-rc1/examples/libpcap/test-captures").
%-define(MY_OS, "Solaris").
%-define(IFACE_EXIST, "pcn0").
%-define(IFACE_NOTEXIST, "blar4").
%% Path to the "test-captures" directory
%-define(TEST_CAPTURES_DIR,
%	"/user/fritchie/edtk-1.0-rc2/examples/libpcap/test-captures").

regression() ->
    {ok, Port1} = ?DRV:start(),
    {ok, 1} = ?DRV:debug(Port1, 1),
    r1(Port1),

io:format("QQQ\n\n\n"),
    {ok, Port2} = ?DRV:start_pipe(),
    {ok, 1} = ?DRV:debug(Port2, 1),
    r1(Port2),

    io:format("All regression tests PASSED.\n"),
    ok.

r1(P) ->
    ok = ?DRV:null(P),

    {ok, PC1} = ?DRV:pcap_open_offline(P, ?TEST_CAPTURES_DIR ++ "/tcpdump.1"),
    {ok, {{Pkt1_TS_s, Pkt1_TS_us}, Pkt1}} = ?DRV:pcap_next(P, PC1),
    {ok, {{Pkt2_TS_s, Pkt2_TS_us}, Pkt2}} = ?DRV:pcap_next(P, PC1),
    {ok, {{Pkt3_TS_s, Pkt3_TS_us}, Pkt3}} = ?DRV:pcap_next(P, PC1),
    {ok, {{Pkt4_TS_s, Pkt4_TS_us}, Pkt4}} = ?DRV:pcap_next(P, PC1),
    {ok, {{Pkt5_TS_s, Pkt5_TS_us}, Pkt5}} = ?DRV:pcap_next(P, PC1),
    {error, {}} = ?DRV:pcap_next(P, PC1),
    {ok, _} = libpcap_drv:pcap_is_swapped(P, PC1),
    {ok, _} = ?DRV:pcap_major_version(P, PC1),
    {ok, _} = ?DRV:pcap_minor_version(P, PC1),
    ok = ?DRV:pcap_close(P, PC1),
    {error, badarg} = ?DRV:pcap_close(P, PC1),

    {ok, Iface1} = ?DRV:pcap_lookupdev(P),
    
    %% XXX On my test platform, ?IFACE_NOTEXIST does not exist, "rl0" does.
    {error, _} = ?DRV:pcap_lookupnet(P, ?IFACE_NOTEXIST),
    {ok, {ExistAddr, ExistMask}} = ?DRV:pcap_lookupnet(P, ?IFACE_EXIST),

    {ok, _BigList} = ?DRV:pcap_findalldevs(P),

    %% Intentional leaking here
    {ok, PC98} = ?DRV:pcap_open_offline(P, ?TEST_CAPTURES_DIR ++ "/tcpdump.1"),
    {ok, PC99} = ?DRV:pcap_open_offline(P, ?TEST_CAPTURES_DIR ++ "/tcpdump.1"),

    {error, _} = ?DRV:pcap_open_live(P, ?IFACE_NOTEXIST, 1514, 1, 2000),
    {ok, PC2} = ?DRV:pcap_open_live(P, ?IFACE_EXIST, 1514, 1, 120 * 1000),
    case ?MY_OS of
	"FreeBSD" ->
	    {ok, 0} = ?DRV:my_set_immediate(P, PC2, 0),
	    {ok, 0} = ?DRV:my_set_immediate(P, PC2, 1);
	_ ->
	    ok
    end,
    ReadPkts = 5,
    io:format("\n\n\nPlease generate at least ~w packets of network traffic\n\n",
	      [ReadPkts]),
    ok = read_packets(P, PC2, ReadPkts),
    io:format("\nGot ~w packets.  Thank you!\n\n", [ReadPkts]),
    {ok, 0} = ?DRV:pcap_getnonblock(P, PC2),
    {ok, 0} = ?DRV:pcap_setnonblock(P, PC2, 1),
    {ok, 1} = ?DRV:pcap_getnonblock(P, PC2),
    
    %% QQQ My "rl0" is an Ethernet device = 1
    {ok, 1} = libpcap_drv:pcap_datalink(P, PC2),
    {ok, 1514} = libpcap_drv:pcap_snapshot(P, PC2),
    {ok, {PS_Recv, PS_Drop}} = libpcap_drv:pcap_stats(P, PC2),
    if
	PS_Recv < ReadPkts -> throw({failed, only_read, PS_Recv});	    
	PS_Drop > 0 -> throw({failed, dropped, PS_Drop});
	true -> ok
    end,

    %% The pcap_stats test must be done before setting this filter
    {error, _} = ?DRV:my_pcap_compile(P, PC2, "BOGUS", 1, ExistMask),
    {ok, _} = ?DRV:pcap_geterr(P, PC2),
    {ok, BPF1} = ?DRV:my_pcap_compile(P, PC2, "arp or icmp", 1, ExistMask),
    {ok, 0} = libpcap_drv:pcap_setfilter(P, PC2, BPF1),
    ok = libpcap_drv:my_pcap_freecode(P, BPF1),
    {error, badarg} = libpcap_drv:my_pcap_freecode(P, BPF1),

    ok = ?DRV:pcap_close(P, PC1),
    
    ok.

read_packets(P, PC, 0) when port(P) ->
    ok;
read_packets(P, PC, N) when port(P) ->
    case ?DRV:pcap_next(P, PC) of
	{ok, {_Pkt_TS, Pkt}} ->
	    io:format("Got packet ~w, ~w bytes\n", [N, size(Pkt)]),
	    read_packets(P, PC, N - 1);
	{error, {}} ->
	    read_packets(P, PC, N);
	Err ->
	    io:format("read_packets got Err =  ~w\n", [Err])
    end.
	    
%{ok, P} = libpcap_drv:start().
%libpcap_drv:pcap_findalldevs(P).
%{ok,[{pcap_if_t,"rl0",
%                [],
%                [{pcap_addr_t,{sockaddr,
%                                  18,
%                                  <<1,0,6,3,6,0,114,108,48,0,48,27,7,128,...>>}, 
%                              {},
%                              {},
%                              {}},
%                 {pcap_addr_t,{sockaddr,
%                                  2,
%                                  <<0,0,10,1,1,1,0,0,0,0,0,0,0,...>>},
%                              {sockaddr,
%                                  2,
%                                  <<0,0,255,255,255>>},
%                              {sockaddr,
%                                  2,
%                                  <<0,0,10,1,1,255,0,0,0,0,...>>},
%                              {}},
%                 {pcap_addr_t,{sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,254,128,0,1,0,0,...>>},
%                              {sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,255,255,255,255,...>>},
%                              {},
%                              {}}],
%                0},
%     {pcap_if_t,"lo0",
%                [],
%                [{pcap_addr_t,{sockaddr,
%                                  18,
%                                  <<4,0,24,3,0,0,108,111,48,0,0,0,0,...>>},
%                              {},
%                              {},
%                              {}},
%                 {pcap_addr_t,{sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,0,0,0,0,0,0,...>>},
%                              {sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,255,255,255,255,...>>},
%                              {},
%                              {}},
%                 {pcap_addr_t,{sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,254,128,0,4,0,...>>},
%                              {sockaddr,
%                                  28,
%                                  <<0,0,0,0,0,0,255,255,255,...>>},
%                              {},
%                              {}},
%                 {pcap_addr_t,{sockaddr,
%                                  2,
%                                  <<0,0,127,0,0,1,0,0,0,0,...>>},
%                              {sockaddr,
%                                  2,
%                                  <<0,0,255>>},
%                              {},
%                              {}}],
%                1}]}

